/*
**                TTempSmooth v0.9.4 for AviSynth 2.5.x
**
**   TTempSmooth is a basic, motion adaptive, temporal smoothing filter. 
**   It currently supports YV12 and YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TTempSmooth.h"

PVideoFrame __stdcall TTempSmooth::GetFrame(int n, IScriptEnvironment *env)
{
	checkBuffer(n, env);
	if (useDiff) 
	{
		if (usePF) filterFrame_PF(n);
		else filterFrame(n);
	}
	else 
	{
		if (usePF) filterFrameNoDiff_PF(n);
		else filterFrameNoDiff(n);
	}
	PVideoFrame dst = env->NewVideoFrame(vi);
	if (vi.IsYV12()) YUVtoYV12(dst);
	else YUVtoYUY2(dst);
	return dst;
}

void TTempSmooth::filterFrame(int n)
{
	unsigned char* frmp = getPointer(getPos(-1, false));
	unsigned char* bufp = getPointer(getPos(maxr, false));
	unsigned char *bufpT1, *bufpT2;
	for (int i=0; i<diameter; ++i) posTable[i] = bufWidthA*getPos2(i);
	int height = vi.height;
	int width = bufWidth;
	int minsc, maxsc;
	getSceneLimits(n, minsc, maxsc);
	int x, y, b, v;
	unsigned int weightY, weightU, weightV;
	unsigned int weightsY, weightsU, weightsV;
	unsigned int sumY, sumU, sumV;
	int cY, cU, cV;
	int tY, tU, tV;
	for (y=0; y<height; ++y)
	{
		for (x=0; x<width; x+=3)
		{
			cY = bufp[x];
			cU = bufp[x+1];
			cV = bufp[x+2];
			weightsY = weightsU = weightsV = cw;
			sumY = cY*cw;
			sumU = cU*cw;
			sumV = cV*cw;
			b = maxr-1;
			if (b>minsc)
			{
				bufpT1 = bufp+posTable[b];
				tY = bufpT1[x];
				tU = bufpT1[x+1];
				tV = bufpT1[x+2];
				if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && abs(cV-tV) < cthresh)
				{
					weightY = weightTY[abs(cY-tY)];
					weightU = weightTC[abs(cU-tU)];
					weightV = weightTC[abs(cV-tV)];
					weightsY += weightY;
					weightsU += weightU;
					weightsV += weightV;
					sumY += tY*weightY;
					sumU += tU*weightU;
					sumV += tV*weightV;
					--b;
					v = 256;
					while(b>minsc) // backwards
					{
						bufpT2 = bufpT1;
						bufpT1 = bufp+posTable[b];
						tY = bufpT1[x];
						tU = bufpT1[x+1];
						tV = bufpT1[x+2];
						if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && 
							abs(cV-tV) < cthresh && abs(tY-bufpT2[x]) < lthresh && 
							abs(tU-bufpT2[x+1]) < cthresh && abs(tV-bufpT2[x+2]) < cthresh)
						{
							weightY = weightTY[abs(cY-tY)+v];
							weightU = weightTC[abs(cU-tU)+v];
							weightV = weightTC[abs(cV-tV)+v];
							weightsY += weightY;
							weightsU += weightU;
							weightsV += weightV;
							sumY += tY*weightY;
							sumU += tU*weightU;
							sumV += tV*weightV;
							--b;
							v += 256;
						}
						else break;
					}
				}
			}
			b = maxr+1;
			if (b<maxsc)
			{
				bufpT1 = bufp+posTable[b];
				tY = bufpT1[x];
				tU = bufpT1[x+1];
				tV = bufpT1[x+2];
				if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && abs(cV-tV) < cthresh)
				{
					weightY = weightTY[abs(cY-tY)];
					weightU = weightTC[abs(cU-tU)];
					weightV = weightTC[abs(cV-tV)];
					weightsY += weightY;
					weightsU += weightU;
					weightsV += weightV;
					sumY += tY*weightY;
					sumU += tU*weightU;
					sumV += tV*weightV;
					++b;
					v = 256;
					while(b<maxsc) // forwards
					{
						bufpT2 = bufpT1;
						bufpT1 = bufp+posTable[b];
						tY = bufpT1[x];
						tU = bufpT1[x+1];
						tV = bufpT1[x+2];
						if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && 
							abs(cV-tV) < cthresh && abs(tY-bufpT2[x]) < lthresh && 
							abs(tU-bufpT2[x+1]) < cthresh && abs(tV-bufpT2[x+2]) < cthresh)
						{
							weightY = weightTY[abs(cY-tY)+v];
							weightU = weightTC[abs(cU-tU)+v];
							weightV = weightTC[abs(cV-tV)+v];
							weightsY += weightY;
							weightsU += weightU;
							weightsV += weightV;
							sumY += tY*weightY;
							sumU += tU*weightU;
							sumV += tV*weightV;
							++b;
							v += 256;
						}
						else break;
					}
				}
			}
			if (fp) // fixed point
			{
				if (!vis_blur)
				{
					frmp[x] = ((cY*(65536-weightsY)) + sumY + 32768) >> 16;
					frmp[x+1] = ((cU*(65536-weightsU)) + sumU + 32768) >> 16;
					frmp[x+2] = ((cV*(65536-weightsV)) + sumV + 32768) >> 16;
				}
				else
				{
					int maxweight = 65536-cw;
					frmp[x] = maxweight == 0 ? 0 : min(((weightsY-cw)*255)/maxweight,255);
					frmp[x+1] = maxweight == 0 ? 0 : min(((weightsU-cw)*255)/maxweight,255);
					frmp[x+2] = maxweight == 0 ? 0 : min(((weightsV-cw)*255)/maxweight,255);
				}
			}
			else // regular average
			{
				if (!vis_blur)
				{
					frmp[x] = (sumY+(weightsY>>1))/weightsY;
					frmp[x+1] = (sumU+(weightsU>>1))/weightsU;
					frmp[x+2] = (sumV+(weightsV>>1))/weightsV;
				}
				else
				{
					frmp[x] = min(((weightsY-cw)*255)/weightsY,255);
					frmp[x+1] = min(((weightsU-cw)*255)/weightsU,255);
					frmp[x+2] = min(((weightsV-cw)*255)/weightsV,255);
				}
			}
		}
		frmp += bufPitch;
		bufp += bufPitch;
	}
}

void TTempSmooth::filterFrame_PF(int n)
{
	unsigned char* frmp = getPointer(getPos(-1, false));
	unsigned char* bufp = getPointer(getPos(maxr, false));
	unsigned char* bufpf = getPointer(getPos(maxr, true));
	unsigned char *bufpT1, *bufpT2;
	for (int i=0; i<diameter; ++i) posTable[i] = bufWidthA*getPos2(i);
	int height = vi.height;
	int width = bufWidth;
	int minsc, maxsc;
	getSceneLimits(n, minsc, maxsc);
	int x, y, b, v;
	unsigned int weightY, weightU, weightV;
	unsigned int weightsY, weightsU, weightsV;
	unsigned int sumY, sumU, sumV;
	int cY, cU, cV;
	int tY, tU, tV;
	for (y=0; y<height; ++y)
	{
		for (x=0; x<width; x+=3)
		{
			cY = bufpf[x];
			cU = bufpf[x+1];
			cV = bufpf[x+2];
			weightsY = weightsU = weightsV = cw;
			sumY = bufp[x]*cw;
			sumU = bufp[x+1]*cw;
			sumV = bufp[x+2]*cw;
			b = maxr-1;
			if (b>minsc)
			{
				bufpT1 = bufpf+posTable[b];
				tY = bufpT1[x];
				tU = bufpT1[x+1];
				tV = bufpT1[x+2];
				if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && abs(cV-tV) < cthresh)
				{
					weightY = weightTY[abs(cY-tY)];
					weightU = weightTC[abs(cU-tU)];
					weightV = weightTC[abs(cV-tV)];
					weightsY += weightY;
					weightsU += weightU;
					weightsV += weightV;
					sumY += bufpT1[x-bufWidthA]*weightY;
					sumU += bufpT1[x+1-bufWidthA]*weightU;
					sumV += bufpT1[x+2-bufWidthA]*weightV;
					--b;
					v = 256;
					while(b>minsc) // backwards
					{
						bufpT2 = bufpT1;
						bufpT1 = bufpf+posTable[b];
						tY = bufpT1[x];
						tU = bufpT1[x+1];
						tV = bufpT1[x+2];
						if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && 
							abs(cV-tV) < cthresh && abs(tY-bufpT2[x]) < lthresh && 
							abs(tU-bufpT2[x+1]) < cthresh && abs(tV-bufpT2[x+2]) < cthresh)
						{
							weightY = weightTY[abs(cY-tY)+v];
							weightU = weightTC[abs(cU-tU)+v];
							weightV = weightTC[abs(cV-tV)+v];
							weightsY += weightY;
							weightsU += weightU;
							weightsV += weightV;
							sumY += bufpT1[x-bufWidthA]*weightY;
							sumU += bufpT1[x+1-bufWidthA]*weightU;
							sumV += bufpT1[x+2-bufWidthA]*weightV;
							--b;
							v += 256;
						}
						else break;
					}
				}
			}
			b = maxr+1;
			if (b<maxsc)
			{
				bufpT1 = bufpf+posTable[b];
				tY = bufpT1[x];
				tU = bufpT1[x+1];
				tV = bufpT1[x+2];
				if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && abs(cV-tV) < cthresh)
				{
					weightY = weightTY[abs(cY-tY)];
					weightU = weightTC[abs(cU-tU)];
					weightV = weightTC[abs(cV-tV)];
					weightsY += weightY;
					weightsU += weightU;
					weightsV += weightV;
					sumY += bufpT1[x-bufWidthA]*weightY;
					sumU += bufpT1[x+1-bufWidthA]*weightU;
					sumV += bufpT1[x+2-bufWidthA]*weightV;
					++b;
					v = 256;
					while(b<maxsc) // forwards
					{
						bufpT2 = bufpT1;
						bufpT1 = bufpf+posTable[b];
						tY = bufpT1[x];
						tU = bufpT1[x+1];
						tV = bufpT1[x+2];
						if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && 
							abs(cV-tV) < cthresh && abs(tY-bufpT2[x]) < lthresh && 
							abs(tU-bufpT2[x+1]) < cthresh && abs(tV-bufpT2[x+2]) < cthresh)
						{
							weightY = weightTY[abs(cY-tY)+v];
							weightU = weightTC[abs(cU-tU)+v];
							weightV = weightTC[abs(cV-tV)+v];
							weightsY += weightY;
							weightsU += weightU;
							weightsV += weightV;
							sumY += bufpT1[x-bufWidthA]*weightY;
							sumU += bufpT1[x+1-bufWidthA]*weightU;
							sumV += bufpT1[x+2-bufWidthA]*weightV;
							++b;
							v += 256;
						}
						else break;
					}
				}
			}
			if (fp) // fixed point
			{
				if (!vis_blur)
				{
					frmp[x] = ((bufp[x]*(65536-weightsY)) + sumY + 32768) >> 16;
					frmp[x+1] = ((bufp[x+1]*(65536-weightsU)) + sumU + 32768) >> 16;
					frmp[x+2] = ((bufp[x+2]*(65536-weightsV)) + sumV + 32768) >> 16;
				}
				else
				{
					int maxweight = 65536-cw;
					frmp[x] = maxweight == 0 ? 0 : min(((weightsY-cw)*255)/maxweight,255);
					frmp[x+1] = maxweight == 0 ? 0 : min(((weightsU-cw)*255)/maxweight,255);
					frmp[x+2] = maxweight == 0 ? 0 : min(((weightsV-cw)*255)/maxweight,255);
				}
			}
			else // regular average
			{
				if (!vis_blur)
				{
					frmp[x] = (sumY+(weightsY>>1))/weightsY;
					frmp[x+1] = (sumU+(weightsU>>1))/weightsU;
					frmp[x+2] = (sumV+(weightsV>>1))/weightsV;
				}
				else
				{
					frmp[x] = min(((weightsY-cw)*255)/weightsY,255);
					frmp[x+1] = min(((weightsU-cw)*255)/weightsU,255);
					frmp[x+2] = min(((weightsV-cw)*255)/weightsV,255);
				}
			}
		}
		frmp += bufPitch;
		bufp += bufPitch;
		bufpf += bufPitch;
	}
}

void TTempSmooth::filterFrameNoDiff(int n)
{
	unsigned char* frmp = getPointer(getPos(-1, false));
	unsigned char* bufp = getPointer(getPos(maxr, false));
	unsigned char* bufpT1, *bufpT2;
	for (int i=0; i<diameter; ++i) posTable[i] = bufWidthA*getPos2(i);
	int height = vi.height;
	int width = bufWidth;
	int minsc, maxsc;
	getSceneLimits(n, minsc, maxsc);
	int x, y, b;
	unsigned int weight, weights;
	unsigned int sumY, sumU, sumV;
	unsigned int invWeight;
	int cY, cU, cV;
	int tY, tU, tV;
	for (y=0; y<height; ++y)
	{
		for (x=0; x<width; x+=3)
		{
			cY = bufp[x];
			cU = bufp[x+1];
			cV = bufp[x+2];
			weights = cw;
			sumY = cY*cw;
			sumU = cU*cw;
			sumV = cV*cw;
			b = maxr-1;
			if (b>minsc)
			{
				bufpT1 = bufp+posTable[b];
				tY = bufpT1[x];
				tU = bufpT1[x+1];
				tV = bufpT1[x+2];
				if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && abs(cV-tV) < cthresh)
				{
					weight = weightTY[b];
					weights += weight;
					sumY += tY*weight;
					sumU += tU*weight;
					sumV += tV*weight;
					--b;
					while(b>minsc) // backwards
					{
						bufpT2 = bufpT1;
						bufpT1 = bufp+posTable[b];
						tY = bufpT1[x];
						tU = bufpT1[x+1];
						tV = bufpT1[x+2];
						if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && 
							abs(cV-tV) < cthresh && abs(tY-bufpT2[x]) < lthresh && 
							abs(tU-bufpT2[x+1]) < cthresh && abs(tV-bufpT2[x+2]) < cthresh)
						{
							weight = weightTY[b];
							weights += weight;
							sumY += tY*weight;
							sumU += tU*weight;
							sumV += tV*weight;
							--b;
						}
						else break;
					}
				}
			}
			b = maxr+1;
			if (b<maxsc)
			{
				bufpT1 = bufp+posTable[b];
				tY = bufpT1[x];
				tU = bufpT1[x+1];
				tV = bufpT1[x+2];
				if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && abs(cV-tV) < cthresh)
				{
					weight = weightTY[b];
					weights += weight;
					sumY += tY*weight;
					sumU += tU*weight;
					sumV += tV*weight;
					++b;
					while(b<maxsc) // forwards
					{
						bufpT2 = bufpT1;
						bufpT1 = bufp+posTable[b];
						tY = bufpT1[x];
						tU = bufpT1[x+1];
						tV = bufpT1[x+2];
						if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && 
							abs(cV-tV) < cthresh && abs(tY-bufpT2[x]) < lthresh && 
							abs(tU-bufpT2[x+1]) < cthresh && abs(tV-bufpT2[x+2]) < cthresh)
						{
							weight = weightTY[b];
							weights += weight;
							sumY += tY*weight;
							sumU += tU*weight;
							sumV += tV*weight;
							++b;
						}
						else break;
					}
				}
			}
			if (fp) // fixed point
			{
				if (!vis_blur)
				{
					invWeight = 65536 - weights;
					frmp[x] = ((cY*invWeight) + sumY + 32768) >> 16;
					frmp[x+1] = ((cU*invWeight) + sumU + 32768) >> 16;
					frmp[x+2] = ((cV*invWeight) + sumV + 32768) >> 16;
				}
				else
				{
					int maxweight = 65536-cw;
					frmp[x] = frmp[x+1] = frmp[x+2] =
						maxweight == 0 ? 0 : min(((weights-cw)*255)/maxweight,255);
				}
			}
			else // regular average
			{
				if (!vis_blur)
				{
					frmp[x] = (sumY+(weights>>1))/weights;
					frmp[x+1] = (sumU+(weights>>1))/weights;
					frmp[x+2] = (sumV+(weights>>1))/weights;
				}
				else
				{
					frmp[x] = frmp[x+1] = frmp[x+2] = 
						min(((weights-cw)*255)/weights,255);
				}
			}
		}
		frmp += bufPitch;
		bufp += bufPitch;
	}
}

void TTempSmooth::filterFrameNoDiff_PF(int n)
{
	unsigned char* frmp = getPointer(getPos(-1, false));
	unsigned char* bufp = getPointer(getPos(maxr, false));
	unsigned char* bufpf = getPointer(getPos(maxr, true));
	unsigned char* bufpT1, *bufpT2;
	for (int i=0; i<diameter; ++i) posTable[i] = bufWidthA*getPos2(i);
	int height = vi.height;
	int width = bufWidth;
	int minsc, maxsc;
	getSceneLimits(n, minsc, maxsc);
	int x, y, b;
	unsigned int weight, weights;
	unsigned int sumY, sumU, sumV;
	unsigned int invWeight;
	int cY, cU, cV;
	int tY, tU, tV;
	for (y=0; y<height; ++y)
	{
		for (x=0; x<width; x+=3)
		{
			cY = bufpf[x];
			cU = bufpf[x+1];
			cV = bufpf[x+2];
			weights = cw;
			sumY = bufp[x]*cw;
			sumU = bufp[x+1]*cw;
			sumV = bufp[x+2]*cw;
			b = maxr-1;
			if (b>minsc)
			{
				bufpT1 = bufpf+posTable[b];
				tY = bufpT1[x];
				tU = bufpT1[x+1];
				tV = bufpT1[x+2];
				if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && abs(cV-tV) < cthresh)
				{
					weight = weightTY[b];
					weights += weight;
					sumY += bufpT1[x-bufWidthA]*weight;
					sumU += bufpT1[x+1-bufWidthA]*weight;
					sumV += bufpT1[x+2-bufWidthA]*weight;
					--b;
					while(b>minsc) // backwards
					{
						bufpT2 = bufpT1;
						bufpT1 = bufpf+posTable[b];
						tY = bufpT1[x];
						tU = bufpT1[x+1];
						tV = bufpT1[x+2];
						if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && 
							abs(cV-tV) < cthresh && abs(tY-bufpT2[x]) < lthresh && 
							abs(tU-bufpT2[x+1]) < cthresh && abs(tV-bufpT2[x+2]) < cthresh)
						{
							weight = weightTY[b];
							weights += weight;
							sumY += bufpT1[x-bufWidthA]*weight;
							sumU += bufpT1[x+1-bufWidthA]*weight;
							sumV += bufpT1[x+2-bufWidthA]*weight;
							--b;
						}
						else break;
					}
				}
			}
			b = maxr+1;
			if (b<maxsc)
			{
				bufpT1 = bufpf+posTable[b];
				tY = bufpT1[x];
				tU = bufpT1[x+1];
				tV = bufpT1[x+2];
				if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && abs(cV-tV) < cthresh)
				{
					weight = weightTY[b];
					weights += weight;
					sumY += bufpT1[x-bufWidthA]*weight;
					sumU += bufpT1[x+1-bufWidthA]*weight;
					sumV += bufpT1[x+2-bufWidthA]*weight;
					++b;
					while(b<maxsc) // forwards
					{
						bufpT2 = bufpT1;
						bufpT1 = bufpf+posTable[b];
						tY = bufpT1[x];
						tU = bufpT1[x+1];
						tV = bufpT1[x+2];
						if (abs(cY-tY) < lthresh && abs(cU-tU) < cthresh && 
							abs(cV-tV) < cthresh && abs(tY-bufpT2[x]) < lthresh && 
							abs(tU-bufpT2[x+1]) < cthresh && abs(tV-bufpT2[x+2]) < cthresh)
						{
							weight = weightTY[b];
							weights += weight;
							sumY += bufpT1[x-bufWidthA]*weight;
							sumU += bufpT1[x+1-bufWidthA]*weight;
							sumV += bufpT1[x+2-bufWidthA]*weight;
							++b;
						}
						else break;
					}
				}
			}
			if (fp) // fixed point
			{
				if (!vis_blur)
				{
					invWeight = 65536 - weights;
					frmp[x] = ((bufp[x]*invWeight) + sumY + 32768) >> 16;
					frmp[x+1] = ((bufp[x+1]*invWeight) + sumU + 32768) >> 16;
					frmp[x+2] = ((bufp[x+2]*invWeight) + sumV + 32768) >> 16;
				}
				else
				{
					int maxweight = 65536-cw;
					frmp[x] = frmp[x+1] = frmp[x+2] =
						maxweight == 0 ? 0 : min(((weights-cw)*255)/maxweight,255);
				}
			}
			else // regular average
			{
				if (!vis_blur)
				{
					frmp[x] = (sumY+(weights>>1))/weights;
					frmp[x+1] = (sumU+(weights>>1))/weights;
					frmp[x+2] = (sumV+(weights>>1))/weights;
				}
				else
				{
					frmp[x] = frmp[x+1] = frmp[x+2] = 
						min(((weights-cw)*255)/weights,255);
				}
			}
		}
		frmp += bufPitch;
		bufp += bufPitch;
		bufpf += bufPitch;
	}
}

void TTempSmooth::getSceneLimits(int n, int &min, int &max)
{
	int b, i;
	min = -1;
	for (b=maxr-1, i=n; i>n-maxr; --i, --b)
	{
		if (sceneTrack[mapn(i)] == 1) { min = b; break; }
	}
	max = diameter;
	for (b=maxr+1, i=n+1; i<=n+maxr; ++i, ++b)
	{
		if (sceneTrack[mapn(i)] == 1) { max = b; break; }
	}
}

void TTempSmooth::checkBuffer(int n, IScriptEnvironment *env)
{
	int b, i;
	if (bufTrack[buf_pos_new] != n-maxr && bufTrack[buf_pos_new] != -20)
	{
		for (b=n-maxr; b<=n+maxr; ++b)
		{
			for (i=0; i<diameter; ++i)
			{
				if (bufTrack[i] == b)
				{
					buf_pos = i-b+n-maxr;
					if (buf_pos < 0) buf_pos += diameter;
					goto found;
				}
			}
		}
		buf_pos = buf_pos_new;
	}
	else buf_pos = buf_pos_new;
found:
	PVideoFrame src1, src2;
	for (b=getPosA(0), i=n-maxr; i<=n+maxr; ++i, ++b)
	{
		if (b>=diameter) b -= diameter;
		if (bufTrack[b] != i)
		{
			if (sceneTrack[mapn(i)] < 0 && scthresh > 0.0f) 
				src2 = usePF ? pfclip->GetFrame(mapn(i-1), env) : child->GetFrame(mapn(i-1), env);
			src1 = usePF ? pfclip->GetFrame(mapn(i), env) : child->GetFrame(mapn(i), env);
			if (sceneTrack[mapn(i)] < 0)
			{
				if (scthresh <= 0.0f) sceneTrack[mapn(i)] = 0;
				else
				{
					sceneTrack[mapn(i)] = detectSceneChange(src1, src2, scthresh, mapn(i), env) ? 1 : 0;
					if (debug && sceneTrack[mapn(i)] == 1)
					{
						sprintf(buf,"TTempSmooth:  frame %d  -  SCENECHANGE DETECTED!!\n", i);
						OutputDebugString(buf);
					}
				}
			}
			if (!usePF)
			{
				if (vi.IsYV12()) insertFrameYV12(src1, b+1);
				else insertFrameYUY2(src1, b+1);
			}
			else
			{
				if (vi.IsYV12()) 
				{
					insertFrameYV12(child->GetFrame(mapn(i), env), b*2+1);
					insertFrameYV12(src1, b*2+2);
				}
				else 
				{
					insertFrameYUY2(child->GetFrame(mapn(i), env), b*2+1);
					insertFrameYUY2(src1, b*2+2);
				}
			}
			bufTrack[b] = i;
		}
	}
	buf_pos_new = buf_pos+1;
	if (buf_pos_new >= diameter) buf_pos_new -= diameter;
}

int TTempSmooth::getPos(int pos, bool pfd)
{
	if (pos < 0) return 0;
	if (!usePF) return ((buf_pos+pos)%diameter)+1;
	return ((buf_pos+pos)%diameter)*2+1 + (pfd ? 1 : 0);
}

int TTempSmooth::getPosA(int pos)
{
	if (pos < 0) return 0;
	return ((buf_pos+pos)%diameter);
}

int TTempSmooth::getPos2(int pos)
{
	return getPos(pos,false)-getPos(maxr,false);
}

int TTempSmooth::mapn(int n)
{
	if (n < 0) return 0;
	if (n > nfrms) return nfrms;
	return n;
}

unsigned char* TTempSmooth::getPointer(int pos)
{
	return bufp_saved+bufWidthA*pos;
}

bool TTempSmooth::detectSceneChange(PVideoFrame &src1, PVideoFrame &src2, double thresh, 
		int n, IScriptEnvironment *env)
{
	unsigned long diff = 0;
	if (env->GetCPUFlags()&CPUF_INTEGER_SSE)
	{
		if (vi.IsYUY2())
			sceneChangeYUY2_ISSE(src1->GetReadPtr(),src2->GetReadPtr(),((src1->GetRowSize()>>4)<<4),
				src1->GetHeight(),src1->GetPitch(),src2->GetPitch(),diff);
		else
			sceneChangeYV12_ISSE(src1->GetReadPtr(PLANAR_Y),src2->GetReadPtr(PLANAR_Y),
				((src1->GetRowSize(PLANAR_Y)>>4)<<4),src1->GetHeight(PLANAR_Y),src1->GetPitch(PLANAR_Y),
				src2->GetPitch(PLANAR_Y),diff);
	}
	else if (env->GetCPUFlags()&CPUF_MMX)
	{
		if (vi.IsYUY2())
			sceneChangeYUY2_MMX(src1->GetReadPtr(),src2->GetReadPtr(),((src1->GetRowSize()>>4)<<4),
				src1->GetHeight(),src1->GetPitch(),src2->GetPitch(),diff);
		else
			sceneChangeYV12_MMX(src1->GetReadPtr(PLANAR_Y),src2->GetReadPtr(PLANAR_Y),
				((src1->GetRowSize(PLANAR_Y)>>4)<<4),src1->GetHeight(PLANAR_Y),src1->GetPitch(PLANAR_Y),
				src2->GetPitch(PLANAR_Y),diff);
	}
	else
	{
		const unsigned char *srcp1 = src1->GetReadPtr(PLANAR_Y);
		const unsigned char *srcp2 = src2->GetReadPtr(PLANAR_Y);
		int height = src1->GetHeight(PLANAR_Y);
		int width = ((src1->GetRowSize(PLANAR_Y)>>4)<<4);
		int pitch1 = src1->GetPitch(PLANAR_Y);
		int pitch2 = src2->GetPitch(PLANAR_Y);
		if (vi.IsYV12())
		{
			for (int y=0; y<height; ++y)
			{
				for (int x=0; x<width; x+=4) 
				{
					diff += abs(srcp1[x+0]-srcp2[x+0]);
					diff += abs(srcp1[x+1]-srcp2[x+1]);
					diff += abs(srcp1[x+2]-srcp2[x+2]);
					diff += abs(srcp1[x+3]-srcp2[x+3]);
				}
				srcp1 += pitch1;
				srcp2 += pitch2;
			}
		}
		else
		{
			for (int y=0; y<height; ++y)
			{
				for (int x=0; x<width; x+=8) 
				{
					diff += abs(srcp1[x+0]-srcp2[x+0]);
					diff += abs(srcp1[x+2]-srcp2[x+2]);
					diff += abs(srcp1[x+4]-srcp2[x+4]);
					diff += abs(srcp1[x+6]-srcp2[x+6]);
				}
				srcp1 += pitch1;
				srcp2 += pitch2;
			}
		}
	}
	if (debug)
	{
		sprintf(buf,"TTempSmooth:  frame %d  -  diff = %u   max = %u  (%f)\n", n, diff,
			unsigned int(thresh*maxdiff/100.0f), double(diff)*100.0f/double(maxdiff));
		OutputDebugString(buf);
	}
	if ((double(diff)*100.0f/double(maxdiff)) > thresh) return true;
	return false;
}

void TTempSmooth::insertFrameYV12(PVideoFrame &src, int pos)
{
	int x, y, v;
	if (!interlaced)
	{
		const unsigned char *srcpY = src->GetReadPtr(PLANAR_Y);
		const unsigned char *srcpU = src->GetReadPtr(PLANAR_U);
		const unsigned char *srcpV = src->GetReadPtr(PLANAR_V);
		int src_pitchY2 = src->GetPitch(PLANAR_Y)<<1;
		int src_pitchUV = src->GetPitch(PLANAR_U);
		int height = src->GetHeight(PLANAR_Y);
		int width = src->GetRowSize(PLANAR_Y);
		const unsigned char *srcpnY = srcpY + (src_pitchY2>>1);
		unsigned char *bufp = getPointer(pos);
		unsigned char *bufpn = bufp + bufPitch;
		int buf_pitch2 = bufPitch<<1;
		for (y=0; y<height; y+=2)
		{
			for (v=0, x=0; x<width; x+=2, v+=6)
			{
				bufp[v] = srcpY[x];
				bufp[v+3] = srcpY[x+1];
				bufpn[v]= srcpnY[x];
				bufpn[v+3] = srcpnY[x+1];
				bufp[v+1] = bufp[v+4] = bufpn[v+1] = bufpn[v+4] = srcpU[x>>1];
				bufp[v+2] = bufp[v+5] = bufpn[v+2] = bufpn[v+5] = srcpV[x>>1];
			}
			srcpY += src_pitchY2;
			srcpnY += src_pitchY2;
			srcpU += src_pitchUV;
			srcpV += src_pitchUV;
			bufp += buf_pitch2;
			bufpn += buf_pitch2;
		}
	}
	else
	{
		const unsigned char *srcpY0 = src->GetReadPtr(PLANAR_Y);
		const unsigned char *srcpU0 = src->GetReadPtr(PLANAR_U);
		const unsigned char *srcpV0 = src->GetReadPtr(PLANAR_V);
		int src_pitchY = src->GetPitch(PLANAR_Y);
		int src_pitchY4 = src_pitchY<<2;
		int src_pitchUV = src->GetPitch(PLANAR_U);
		int src_pitchUV2 = src_pitchUV<<1;
		int height = src->GetHeight(PLANAR_Y);
		int width = src->GetRowSize(PLANAR_Y);
		const unsigned char *srcpY1 = srcpY0 + src_pitchY;
		const unsigned char *srcpY2 = srcpY1 + src_pitchY;
		const unsigned char *srcpY3 = srcpY2 + src_pitchY;
		const unsigned char *srcpU1 = srcpU0 + src_pitchUV;
		const unsigned char *srcpV1 = srcpV0 + src_pitchUV;
		unsigned char *bufp0 = getPointer(pos);
		unsigned char *bufp1 = bufp0 + bufPitch;
		unsigned char *bufp2 = bufp1 + bufPitch;
		unsigned char *bufp3 = bufp2 + bufPitch;
		int buf_pitch4 = bufPitch<<2;
		for (y=0; y<height; y+=4)
		{
			for (v=0, x=0; x<width; x+=2, v+=6)
			{
				bufp0[v] = srcpY0[x];
				bufp0[v+3] = srcpY0[x+1];
				bufp1[v] = srcpY1[x];
				bufp1[v+3] = srcpY1[x+1];
				bufp2[v] = srcpY2[x];
				bufp2[v+3] = srcpY2[x+1];
				bufp3[v] = srcpY3[x];
				bufp3[v+3] = srcpY3[x+1];
				bufp0[v+1] = bufp0[v+4] = bufp2[v+1] = bufp2[v+4] = srcpU0[x>>1];
				bufp0[v+2] = bufp0[v+5] = bufp2[v+2] = bufp2[v+5] = srcpV0[x>>1];
				bufp1[v+1] = bufp1[v+4] = bufp3[v+1] = bufp3[v+4] = srcpU1[x>>1];
				bufp1[v+2] = bufp1[v+5] = bufp3[v+2] = bufp3[v+5] = srcpV1[x>>1];
			}
			srcpY0 += src_pitchY4;
			srcpY1 += src_pitchY4;
			srcpY2 += src_pitchY4;
			srcpY3 += src_pitchY4;
			srcpU0 += src_pitchUV2;
			srcpU1 += src_pitchUV2;
			srcpV0 += src_pitchUV2;
			srcpV1 += src_pitchUV2;
			bufp0 += buf_pitch4;
			bufp1 += buf_pitch4;
			bufp2 += buf_pitch4;
			bufp3 += buf_pitch4;
		}
	}
}

void TTempSmooth::insertFrameYUY2(PVideoFrame &src, int pos)
{
	const unsigned char *srcp = src->GetReadPtr();
	int src_pitch = src->GetPitch();
	int height = src->GetHeight();
	int width = src->GetRowSize();
	unsigned char *bufp = getPointer(pos);
	int buf_pitch = bufPitch, x, y, v;
	for (y=0; y<height; ++y)
	{
		for (v=0, x=0; x<width; x+=4, v+=6)
		{
			bufp[v] = srcp[x];
			bufp[v+1] = bufp[v+4] = srcp[x+1];
			bufp[v+3] = srcp[x+2];
			bufp[v+2] = bufp[v+5] = srcp[x+3];
		}
		srcp += src_pitch;
		bufp += buf_pitch;
	}
}

void TTempSmooth::YUVtoYUY2(PVideoFrame &dst)
{
	unsigned char *dstp = dst->GetWritePtr();
	int height = dst->GetHeight();
	int width = dst->GetRowSize();
	int dst_pitch = dst->GetPitch();
	unsigned char *frmp = getPointer(getPos(-1, false));
	int frm_pitch = bufPitch, x, y, v;
	if (vis_blur == 0)
	{
		for (y=0; y<height; ++y)
		{
			for (v=0, x=0; x<width; x+=4, v+=6)
			{
				dstp[x] = frmp[v];
				dstp[x+1] = (frmp[v+1]+frmp[v+4]+1)>>1;
				dstp[x+2] = frmp[v+3];
				dstp[x+3] = (frmp[v+2]+frmp[v+5]+1)>>1;
			}
			dstp += dst_pitch;
			frmp += frm_pitch;
		}
	}
	else if (vis_blur == 1) // Y
	{
		for (y=0; y<height; ++y)
		{
			for (v=0, x=0; x<width; x+=4, v+=6)
			{
				dstp[x] = frmp[v];
				dstp[x+2] = frmp[v+3];
				dstp[x+1] = dstp[x+3] = 128;
			}
			dstp += dst_pitch;
			frmp += frm_pitch;
		}
	}
	else if (vis_blur == 2) // U
	{
		for (y=0; y<height; ++y)
		{
			for (v=0, x=0; x<width; x+=4, v+=6)
			{
				dstp[x] = frmp[v+1];
				dstp[x+2] = frmp[v+4];
				dstp[x+1] = dstp[x+3] = 128;
			}
			dstp += dst_pitch;
			frmp += frm_pitch;
		}
	}
	else if (vis_blur == 3) // V
	{
		for (y=0; y<height; ++y)
		{
			for (v=0, x=0; x<width; x+=4, v+=6)
			{
				dstp[x] = frmp[v+2];
				dstp[x+2] = frmp[v+5];
				dstp[x+1] = dstp[x+3] = 128;
			}
			dstp += dst_pitch;
			frmp += frm_pitch;
		}
	}
}

void TTempSmooth::YUVtoYV12(PVideoFrame &dst)
{
	unsigned char *dstpY = dst->GetWritePtr(PLANAR_Y);
	unsigned char *dstpU = dst->GetWritePtr(PLANAR_U);
	unsigned char *dstpV = dst->GetWritePtr(PLANAR_V);
	int height = dst->GetHeight(PLANAR_Y);
	int width = dst->GetRowSize(PLANAR_Y);
	int dst_pitchY2 = dst->GetPitch(PLANAR_Y)<<1;
	int dst_pitchUV = dst->GetPitch(PLANAR_U);
	unsigned char *dstpnY = dstpY + (dst_pitchY2>>1);
	unsigned char *frmp = getPointer(getPos(-1, false));
	unsigned char *frmpn = frmp + bufPitch;
	int frm_pitch2 = bufPitch<<1, x, y, v;
	if (!vis_blur)
	{
		if (!interlaced)
		{
			for (y=0; y<height; y+=2)
			{
				for (v=0, x=0; x<width; x+=2, v+=6)
				{
					dstpY[x] = frmp[v];
					dstpY[x+1] = frmp[v+3];
					dstpnY[x] = frmpn[v];
					dstpnY[x+1] = frmpn[v+3];
					dstpU[x>>1] = (frmp[v+1]+frmp[v+4]+frmpn[v+1]+frmpn[v+4]+2)>>2;
					dstpV[x>>1] = (frmp[v+2]+frmp[v+5]+frmpn[v+2]+frmpn[v+5]+2)>>2;
				}
				dstpY += dst_pitchY2;
				dstpnY += dst_pitchY2;
				dstpU += dst_pitchUV;
				dstpV += dst_pitchUV;
				frmp += frm_pitch2;
				frmpn += frm_pitch2;
			}
		}
		else
		{
			unsigned char *dstpY0 = dst->GetWritePtr(PLANAR_Y);
			unsigned char *dstpU0 = dst->GetWritePtr(PLANAR_U);
			unsigned char *dstpV0 = dst->GetWritePtr(PLANAR_V);
			int height = dst->GetHeight(PLANAR_Y);
			int width = dst->GetRowSize(PLANAR_Y);
			int dst_pitchY = dst->GetPitch(PLANAR_Y);
			int dst_pitchY4 = dst_pitchY<<2;
			int dst_pitchUV = dst->GetPitch(PLANAR_U);
			int dst_pitchUV2 = dst_pitchUV<<1;
			unsigned char *dstpY1 = dstpY0 + dst_pitchY;
			unsigned char *dstpY2 = dstpY1 + dst_pitchY;
			unsigned char *dstpY3 = dstpY2 + dst_pitchY;
			unsigned char *dstpU1 = dstpU0 + dst_pitchUV;
			unsigned char *dstpV1 = dstpV0 + dst_pitchUV;
			unsigned char *frmp0 = getPointer(getPos(-1, false));
			unsigned char *frmp1 = frmp0 + bufPitch;
			unsigned char *frmp2 = frmp1 + bufPitch;
			unsigned char *frmp3 = frmp2 + bufPitch;
			int frm_pitch4 = bufPitch<<2;
			for (y=0; y<height; y+=4)
			{
				for (v=0, x=0; x<width; x+=2, v+=6)
				{
					dstpY0[x] = frmp0[v];
					dstpY0[x+1] = frmp0[v+3];
					dstpY1[x] = frmp1[v];
					dstpY1[x+1] = frmp1[v+3];
					dstpY2[x] = frmp2[v];
					dstpY2[x+1] = frmp2[v+3];
					dstpY3[x] = frmp3[v];
					dstpY3[x+1] = frmp3[v+3];
					dstpU0[x>>1] = (frmp0[v+1]+frmp0[v+4]+frmp2[v+1]+frmp2[v+4]+2)>>2;
					dstpV0[x>>1] = (frmp0[v+2]+frmp0[v+5]+frmp2[v+2]+frmp2[v+5]+2)>>2;
					dstpU1[x>>1] = (frmp1[v+1]+frmp1[v+4]+frmp3[v+1]+frmp3[v+4]+2)>>2;
					dstpV1[x>>1] = (frmp1[v+2]+frmp1[v+5]+frmp3[v+2]+frmp3[v+5]+2)>>2;
				}
				dstpY0 += dst_pitchY4;
				dstpY1 += dst_pitchY4;
				dstpY2 += dst_pitchY4;
				dstpY3 += dst_pitchY4;
				dstpU0 += dst_pitchUV2;
				dstpU1 += dst_pitchUV2;
				dstpV0 += dst_pitchUV2;
				dstpV1 += dst_pitchUV2;
				frmp0 += frm_pitch4;
				frmp1 += frm_pitch4;
				frmp2 += frm_pitch4;
				frmp3 += frm_pitch4;
			}
		}
	}
	else if (vis_blur == 1) // Y
	{
		for (y=0; y<height; y+=2)
		{
			for (v=0, x=0; x<width; x+=2, v+=6)
			{
				dstpY[x] = frmp[v];
				dstpY[x+1] = frmp[v+3];
				dstpnY[x] = frmpn[v];
				dstpnY[x+1] = frmpn[v+3];
				dstpU[x>>1] = dstpV[x>>1] = 128;
			}
			dstpY += dst_pitchY2;
			dstpnY += dst_pitchY2;
			dstpU += dst_pitchUV;
			dstpV += dst_pitchUV;
			frmp += frm_pitch2;
			frmpn += frm_pitch2;
		}
	}
	else if (vis_blur == 2) // U
	{
		for (y=0; y<height; y+=2)
		{
			for (v=0, x=0; x<width; x+=2, v+=6)
			{
				dstpY[x] = frmp[v+1];
				dstpY[x+1] = frmp[v+4];
				dstpnY[x] = frmpn[v+1];
				dstpnY[x+1] = frmpn[v+4];
				dstpU[x>>1] = dstpV[x>>1] = 128;
			}
			dstpY += dst_pitchY2;
			dstpnY += dst_pitchY2;
			dstpU += dst_pitchUV;
			dstpV += dst_pitchUV;
			frmp += frm_pitch2;
			frmpn += frm_pitch2;
		}
	}
	else if (vis_blur == 3) // V
	{
		for (y=0; y<height; y+=2)
		{
			for (v=0, x=0; x<width; x+=2, v+=6)
			{
				dstpY[x] = frmp[v+2];
				dstpY[x+1] = frmp[v+5];
				dstpnY[x] = frmpn[v+2];
				dstpnY[x+1] = frmpn[v+5];
				dstpU[x>>1] = dstpV[x>>1] = 128;
			}
			dstpY += dst_pitchY2;
			dstpnY += dst_pitchY2;
			dstpU += dst_pitchUV;
			dstpV += dst_pitchUV;
			frmp += frm_pitch2;
			frmpn += frm_pitch2;
		}
	}
}

void TTempSmooth::buildTables(IScriptEnvironment *env)
{
	if (useDiff)
	{
		weightTY = (unsigned int *)_aligned_malloc(256*maxr*sizeof(unsigned int), 16);
		if (weightTY == NULL) env->ThrowError("TTempSmooth:  aligned malloc failure!");
		weightTC = (unsigned int *)_aligned_malloc(256*maxr*sizeof(unsigned int), 16);
		if (weightTC == NULL) env->ThrowError("TTempSmooth:  aligned malloc failure!");
		double stepsub, basestep, scale;
		double dt[15], rtY[256], rtC[256], sum = 0.0f;
		memset(dt, 0, 15*sizeof(double));
		memset(rtY, 0, 256*sizeof(double));
		memset(rtC, 0, 256*sizeof(double));
		for (int i=0; i<strength && i<=maxr; ++i) dt[i] = 1.0f;
		for (int i=strength; i<=maxr; ++i) dt[i] = 1.0f/double(i-strength+2);
		sum += dt[0];
		for (int i=1; i<=maxr; ++i) sum += dt[i]*2.0f;
		scale = 65536.0f/sum;
		if (lthresh > 0)
		{
			stepsub = 256.0f / ((double)(lthresh-min(lmdiff,lthresh-1)));
			basestep = 256.0f;
			for (int i=0; i<lthresh; ++i)
			{
				if (lmdiff > i) rtY[i] = 256.0f;
				else
				{
					if (basestep > 0.0f) rtY[i] = basestep;
					else break;
					basestep -= stepsub;
				}
			}
		}
		if (cthresh > 0)
		{
			stepsub = 256.0f / ((double)(cthresh-min(cmdiff,cthresh-1)));
			basestep = 256.0f;
			for (int i=0; i<cthresh; ++i)
			{
				if (cmdiff > i) rtC[i] = 256.0f;
				else
				{
					if (basestep > 0.0f) rtC[i] = basestep;
					else break;
					basestep -= stepsub;
				}
			}
		}
		for (int i=1; i<=maxr; ++i)
		{
			for (int v=0; v<256; ++v) 
			{
				weightTY[v+((i-1)<<8)] = unsigned int((dt[i]*scale*rtY[v]/256.0f)+0.5f);
				weightTC[v+((i-1)<<8)] = unsigned int((dt[i]*scale*rtC[v]/256.0f)+0.5f);
			}
		}
		cw = unsigned int(dt[0]*scale+0.5f);
	}
	else
	{
		weightTY = (unsigned int *)_aligned_malloc(diameter*sizeof(unsigned int), 16);
		if (!weightTY) env->ThrowError("TTempSmooth:  malloc failure (weightTY)!");
		double dt[15], sum = 0.0f, scale;
		memset(dt, 0, 15*sizeof(double));
		for (int i=0; i<strength && i<=maxr; ++i) dt[maxr-i] = dt[maxr+i] = 1.0f;
		for (int i=strength; i<=maxr; ++i) dt[maxr-i] = dt[maxr+i] = 1.0f/double(i-strength+2);
		for (int i=0; i<diameter; ++i) sum += dt[i];
		scale = 65536.0f/sum;
		for (int i=0; i<diameter; ++i) weightTY[i] = unsigned int(scale*dt[i]+0.5f);
		cw = weightTY[maxr];
	}
}

void TTempSmooth::allocateBuffer(IScriptEnvironment *env)
{
	bufWidth = 3*vi.width;
	bufWidthA = ((bufWidth+15)>>4)<<4;
	if (!usePF) bufPitch = bufWidthA*(diameter+1);
	else bufPitch = bufWidthA*(diameter*2+1);
	bufBytesA = bufPitch*vi.height;
	bufp_saved = (unsigned char *)_aligned_malloc(bufBytesA*sizeof(unsigned char), 16);
	if (!bufp_saved) env->ThrowError("TTempSmooth:  malloc failure (bufp_saved)!");
	buf_pos = buf_pos_new = 0;
}

TTempSmooth::TTempSmooth(PClip _child, int _maxr, int _lthresh, int _cthresh, int _lmdiff, int _cmdiff,
		int _strength, double _scthresh, bool _fp, int _vis_blur, bool _debug, bool _interlaced, 
		PClip _pfclip, IScriptEnvironment *env) : GenericVideoFilter(_child), maxr(_maxr), lthresh(_lthresh), 
		cthresh(_cthresh), lmdiff(_lmdiff), cmdiff(_cmdiff), strength(_strength), scthresh(_scthresh), 
		fp(_fp), vis_blur(_vis_blur), debug(_debug), interlaced(_interlaced), pfclip(_pfclip)
{
	sceneTrack = bufTrack = NULL;
	bufp_saved = NULL;
	weightTY = weightTC = NULL;
	posTable = NULL;
	if (!vi.IsYV12() && !vi.IsYUY2())
		env->ThrowError("TTempSmooth:  only YV12 and YUY2 input supported!");
	if (maxr < 1 || maxr > 7)
		env->ThrowError("TTempSmooth:  maxr must be set to 1, 2, 3, 4, 5, 6, or 7!");
	if (lthresh < 1 || cthresh < 1 || lthresh > 256 || cthresh > 256)
		env->ThrowError("TTempSmooth:  lthresh and cthresh must be between 1 and 256 (inclusive)!");
	if (lmdiff < 0 || lmdiff > 255 || cmdiff < 0 || cmdiff > 255)
		env->ThrowError("TTempSmooth:  lmdiff and cmdiff must be between 0 and 255 (inclusive)!");
	if (strength < 1 || strength > 8)
		env->ThrowError("TTempSmooth:  strength must be between 1 and 8 (inclusive)!");
	if (vis_blur < 0 || vis_blur > 3)
		env->ThrowError("TTempSmooth:  vis_blur must be set to 0, 1, 2, or 3!");
	if (pfclip)
	{
		VideoInfo vipf = pfclip->GetVideoInfo();
		if (vipf.width != vi.width || vipf.height != vi.height)
			env->ThrowError("TTempSmooth:  pfclip dimensions do not match those of the input clip!");
		if (!vipf.IsYV12() && !vipf.IsYUY2())
			env->ThrowError("TTempSmooth:  pfclip must be in YV12 or YUY2 colorspace!");
		if ((vipf.IsYV12() && vi.IsYUY2()) || (vipf.IsYUY2() && vi.IsYV12()))
			env->ThrowError("TTempSmooth:  pfclip does not have the same colorspace as the input clip!");
		if (vipf.num_frames != vi.num_frames)
			env->ThrowError("TTempSmooth:  pfclip has a different number of frames then the input clip!");
		usePF = true;
		pfclip->SetCacheHints(CACHE_RANGE, 2);
	}
	else usePF = false;
	diameter = maxr*2+1;
	nfrms = vi.num_frames-1;
	useDiff = (lthresh > lmdiff+1 || cthresh > cmdiff+1);
	// Caching is done ourselves, except for scenechange detection for each
	// new frame that is processed... which requires frame n-1
	child->SetCacheHints(CACHE_RANGE, 2);
	buildTables(env);
	allocateBuffer(env);
	bufTrack = (int*)malloc(diameter*sizeof(int));
	if (!bufTrack) env->ThrowError("TTempSmooth:  malloc failure (bufTrack)!");
	for (int i=0; i<diameter; ++i) bufTrack[i] = -20;
	sceneTrack = (int*)malloc(vi.num_frames*sizeof(int));
	if (!sceneTrack) env->ThrowError("TTempSmooth:  malloc failure (sceneTrack)!");
	for (int i=0; i<=nfrms; ++i) sceneTrack[i] = -20;
	maxdiff = (vi.height*((vi.width>>4)<<4)*219);
	posTable = (int*)_aligned_malloc(diameter*sizeof(int),16);
	if (!posTable) env->ThrowError("TTempSmooth:  malloc failure (posTable)!");
	if (debug)
	{
		sprintf(buf,"TTempSmooth %s by tritical\n", VERSION);
		OutputDebugString(buf);
	}
}

TTempSmooth::~TTempSmooth()
{
	if (sceneTrack) free(sceneTrack);
	if (bufTrack) free(bufTrack);
	if (weightTY) _aligned_free(weightTY);
	if (weightTC) _aligned_free(weightTC);
	if (bufp_saved) _aligned_free(bufp_saved);
	if (posTable) _aligned_free(posTable);
}

AVSValue __cdecl Create_TTempSmooth(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new TTempSmooth(args[0].AsClip(),args[1].AsInt(3),args[2].AsInt(4),args[3].AsInt(5),
		args[4].AsInt(2),args[5].AsInt(3),args[6].AsInt(2),args[7].AsFloat(12.0f),
		args[8].AsBool(true),args[9].AsInt(0),args[10].AsBool(false),args[11].AsBool(false),
		args[12].IsClip()?args[12].AsClip():NULL,env);
}

AVSValue __cdecl Create_TTempSmoothF(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new TTempSmooth(args[0].AsClip(),args[1].AsInt(3),args[2].AsInt(4),args[3].AsInt(5),
		255,255,args[4].AsInt(2),args[5].AsFloat(12.0f),args[6].AsBool(true),args[7].AsInt(0),
		args[8].AsBool(false),args[9].AsBool(false),args[10].IsClip()?args[10].AsClip():NULL,env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	env->AddFunction("TTempSmooth", "c[maxr]i[lthresh]i[cthresh]i[lmdiff]i[cmdiff]i[strength]i" \
		"[scthresh]f[fp]b[vis_blur]i[debug]b[interlaced]b[pfclip]c", Create_TTempSmooth, 0);
	env->AddFunction("TTempSmoothF", "c[maxr]i[lthresh]i[cthresh]i[strength]i[scthresh]f[fp]b" \
		"[vis_blur]i[debug]b[interlaced]b[pfclip]c", Create_TTempSmoothF, 0);
	return 0;
}