/*
**                TTempSmooth v0.9.4 for AviSynth 2.5.x
**
**   TTempSmooth is a basic, motion adaptive, temporal smoothing filter. 
**   It currently supports YV12 and YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <windows.h>
#include <stdio.h>
#include <malloc.h>
#include "avisynth.h"

#define VERSION "v0.9.4"

class TTempSmooth : public GenericVideoFilter
{
private:
	int maxr;		// maximum temporal radius
	int lthresh;	// temporal luma threshold
	int cthresh;	// temporal chroma threshold
	int lmdiff;		// maintain diff luma
	int cmdiff;		// maintain diff chroma
	int strength;	// spatial weighting strength
	double scthresh;// scenechange threshold
	bool fp;		// fixed point average
	int vis_blur;	// visualize blur
	bool debug;		// debug output
	bool interlaced;// interlaced yv12 input
	PClip pfclip;	// pre-filtered clip for difference comparisons
	char buf[512];
	bool useDiff, usePF;
	unsigned long maxdiff;
	int diameter, nfrms, *posTable;
	unsigned char *bufp_saved;
	unsigned int *weightTY, *weightTC, cw;
	int *sceneTrack, *bufTrack;
	int bufBytesA, bufWidth, bufWidthA, bufPitch;
	int buf_pos, buf_pos_new;
	void TTempSmooth::YUVtoYV12(PVideoFrame &dst);
	void TTempSmooth::YUVtoYUY2(PVideoFrame &dst);
	void TTempSmooth::insertFrameYUY2(PVideoFrame &src, int pos);
	void TTempSmooth::insertFrameYV12(PVideoFrame &src, int pos);
	bool TTempSmooth::detectSceneChange(PVideoFrame &src1, PVideoFrame &src2, 
		double thresh, int n, IScriptEnvironment *env);
	void TTempSmooth::sceneChangeYV12_ISSE(const unsigned char *prvp, const unsigned char *nxtp, 
		int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff);
	void TTempSmooth::sceneChangeYUY2_ISSE(const unsigned char *prvp, const unsigned char *nxtp, 
		int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff);
	void TTempSmooth::sceneChangeYV12_MMX(const unsigned char *prvp, const unsigned char *nxtp, 
		int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff);
	void TTempSmooth::sceneChangeYUY2_MMX(const unsigned char *prvp, const unsigned char *nxtp, 
		int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff);
	int TTempSmooth::mapn(int n);
	void TTempSmooth::checkBuffer(int n, IScriptEnvironment *env);
	unsigned char* TTempSmooth::getPointer(int pos);
	int TTempSmooth::getPos(int pos, bool pfd);
	int TTempSmooth::getPosA(int pos);
	int TTempSmooth::getPos2(int pos);
	void TTempSmooth::filterFrameNoDiff(int n);
	void TTempSmooth::filterFrameNoDiff_PF(int n);
	void TTempSmooth::filterFrame(int n);
	void TTempSmooth::filterFrame_PF(int n);
	void TTempSmooth::getSceneLimits(int n, int &min, int &max);
	void TTempSmooth::buildTables(IScriptEnvironment *env);
	void TTempSmooth::allocateBuffer(IScriptEnvironment *env);

public:
	PVideoFrame __stdcall TTempSmooth::GetFrame(int n, IScriptEnvironment *env);
	TTempSmooth::TTempSmooth(PClip _child, int _maxr, int _lthresh, int _cthresh, 
		int _lmdiff, int _cmdiff, int _strength, double _scthresh, bool _fp, 
		int _vis_blur, bool _debug, bool _interlaced, PClip _pfclip, IScriptEnvironment *env);
	TTempSmooth::~TTempSmooth();
};