/*
**                TTempSmooth v0.9.4 for AviSynth 2.5.x
**
**   TTempSmooth is a basic, motion adaptive, temporal smoothing filter. 
**   It currently supports YV12 and YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TTempSmooth.h"

__declspec(align(16)) const __int64 lumamask = 0x00FF00FF00FF00FF;

void TTempSmooth::sceneChangeYUY2_ISSE(const unsigned char *prvp, const unsigned char *nxtp, 
	int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff)
{
	if (height&1)
	{
		__asm
		{
			mov edi,prvp
 			mov esi,nxtp
 			mov ecx,width
			mov edx,prv_pitch
			mov ebx,nxt_pitch
			movq mm4,lumamask
			movq mm5,lumamask
			pxor mm6,mm6
			pxor mm7,mm7
yloop1:
			xor eax,eax
 			align 16
xloop1:
			movq mm0,[edi+eax]
			movq mm1,[edi+eax+8]
			movq mm2,[esi+eax]
			movq mm3,[esi+eax+8]
			pand mm0,mm4
			pand mm1,mm5
			pand mm2,mm4
			pand mm3,mm5
			psadbw mm0,mm2
			psadbw mm1,mm3
			paddd mm6,mm0
			paddd mm7,mm1
			add eax,16
			cmp eax,ecx
			jl xloop1
 			add edi,edx
 			add esi,ebx
 			dec height
 			jnz yloop1
			paddd mm6,mm7
			mov eax,diff
			movd [eax],mm6
 			emms
		}
	}
	else
	{
		height >>= 1;
		prv_pitch <<= 1;
		nxt_pitch <<= 1;
		__asm
		{
			mov edi,prvp
 			mov esi,nxtp
 			mov ecx,width
			mov edx,prv_pitch
			mov ebx,nxt_pitch
			shr edx,1
			shr ebx,1
			add edx,edi
			add ebx,esi
			movq mm4,lumamask
			movq mm5,lumamask
			pxor mm6,mm6
			pxor mm7,mm7
yloop2:
			xor eax,eax
 			align 16
xloop2:
			movq mm0,[edi+eax]
			movq mm1,[edi+eax+8]
			movq mm2,[esi+eax]
			movq mm3,[esi+eax+8]
			pand mm0,mm4
			pand mm1,mm5
			pand mm2,mm4
			pand mm3,mm5
			psadbw mm0,mm2
			psadbw mm1,mm3
			paddd mm6,mm0
			paddd mm7,mm1
			movq mm2,[edx+eax]
			movq mm3,[edx+eax+8]
			movq mm0,[ebx+eax]
			movq mm1,[ebx+eax+8]
			pand mm2,mm4
			pand mm3,mm5
			pand mm0,mm4
			pand mm1,mm5
			psadbw mm2,mm0
			psadbw mm3,mm1
			paddd mm6,mm2
			paddd mm7,mm3
			add eax,16
			cmp eax,ecx
			jl xloop2
 			add edi,prv_pitch
 			add esi,nxt_pitch
			add edx,prv_pitch
			add ebx,nxt_pitch
 			dec height
 			jnz yloop2
			paddd mm6,mm7
			mov eax,diff
			movd [eax],mm6
 			emms
		}
	}
}

void TTempSmooth::sceneChangeYUY2_MMX(const unsigned char *prvp, const unsigned char *nxtp, 
	int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff)
{
	__asm
	{
		mov edi,prvp
 		mov esi,nxtp
 		mov ecx,width
		mov edx,prv_pitch
		mov ebx,nxt_pitch
		pxor mm6,mm6
		pxor mm7,mm7
yloop:
		xor eax,eax
 		align 16
xloop:
		movq mm0,[edi+eax]
		movq mm1,[edi+eax+8]
		movq mm2,[esi+eax]
		movq mm3,[esi+eax+8]
		movq mm4,mm0
		movq mm5,mm1
		psubusb mm4,mm2
		psubusb mm5,mm3
		psubusb mm2,mm0
		psubusb mm3,mm1
		por mm2,mm4
		por mm3,mm5
		pxor mm0,mm0
		pxor mm1,mm1
		pand mm2,lumamask
		pand mm3,lumamask
		movq mm4,mm2
		movq mm5,mm3
		punpcklbw mm2,mm0
		punpcklbw mm3,mm1
		punpckhbw mm4,mm0
		punpckhbw mm5,mm1
		paddw mm2,mm3
		paddw mm4,mm5
		paddw mm2,mm4
		movq mm3,mm2
		punpcklwd mm2,mm0
		punpckhwd mm3,mm1
		paddd mm6,mm2
		paddd mm7,mm3
		add eax,16
		cmp eax,ecx
		jl xloop
 		add edi,edx
 		add esi,ebx
 		dec height
 		jnz yloop
		paddd mm6,mm7
		movq mm7,mm6
		psrlq mm6,32
		paddd mm6,mm7
		mov eax,diff
		movd [eax],mm6
 		emms
	}
}

void TTempSmooth::sceneChangeYV12_ISSE(const unsigned char *prvp, const unsigned char *nxtp, 
	int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff)
{
	if (height&1)
	{
		__asm
		{
			mov edi,prvp
 			mov esi,nxtp
 			mov ecx,width
			mov edx,prv_pitch
			mov ebx,nxt_pitch
			pxor mm6,mm6
			pxor mm7,mm7
yloop1:
			xor eax,eax
 			align 16
xloop1:
			movq mm0,[edi+eax]
			movq mm1,[edi+eax+8]
			movq mm2,[esi+eax]
			movq mm3,[esi+eax+8]
			psadbw mm0,mm2
			psadbw mm1,mm3
			paddd mm6,mm0
			paddd mm7,mm1
			add eax,16
			cmp eax,ecx
			jl xloop1
 			add edi,edx
 			add esi,ebx
 			dec height
 			jnz yloop1
			paddd mm6,mm7
			mov eax,diff
			movd [eax],mm6
 			emms
		}
	}
	else
	{
		height >>= 1;
		prv_pitch <<= 1;
		nxt_pitch <<= 1;
		__asm
		{
			mov edi,prvp
 			mov esi,nxtp
 			mov ecx,width
			mov edx,prv_pitch
			mov ebx,nxt_pitch
			shr edx,1
			shr ebx,1
			add edx,edi
			add ebx,esi
			pxor mm6,mm6
			pxor mm7,mm7
yloop2:
			xor eax,eax
 			align 16
xloop2:
			movq mm0,[edi+eax]
			movq mm1,[edi+eax+8]
			movq mm2,[esi+eax]
			movq mm3,[esi+eax+8]
			psadbw mm0,mm2
			psadbw mm1,mm3
			paddd mm6,mm0
			paddd mm7,mm1
			movq mm4,[edx+eax]
			movq mm5,[edx+eax+8]
			movq mm2,[ebx+eax]
			movq mm3,[ebx+eax+8]
			psadbw mm4,mm2
			psadbw mm5,mm3
			paddd mm6,mm4
			paddd mm7,mm5
			add eax,16
			cmp eax,ecx
			jl xloop2
 			add edi,prv_pitch
 			add esi,nxt_pitch
			add edx,prv_pitch
			add ebx,nxt_pitch
 			dec height
 			jnz yloop2
			paddd mm6,mm7
			mov eax,diff
			movd [eax],mm6
 			emms
		}
	}
}

void TTempSmooth::sceneChangeYV12_MMX(const unsigned char *prvp, const unsigned char *nxtp, 
	int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff)
{
	__asm
	{
		mov edi,prvp
 		mov esi,nxtp
 		mov ecx,width
		mov edx,prv_pitch
		mov ebx,nxt_pitch
		pxor mm6,mm6
		pxor mm7,mm7
yloop:
		xor eax,eax
 		align 16
xloop:
		movq mm0,[edi+eax]
		movq mm1,[edi+eax+8]
		movq mm2,[esi+eax]
		movq mm3,[esi+eax+8]
		movq mm4,mm0
		movq mm5,mm1
		psubusb mm4,mm2
		psubusb mm5,mm3
		psubusb mm2,mm0
		psubusb mm3,mm1
		por mm2,mm4
		por mm3,mm5
		pxor mm0,mm0
		pxor mm1,mm1
		movq mm4,mm2
		movq mm5,mm3
		punpcklbw mm2,mm0
		punpcklbw mm3,mm1
		punpckhbw mm4,mm0
		punpckhbw mm5,mm1
		paddw mm2,mm3
		paddw mm4,mm5
		paddw mm2,mm4
		movq mm3,mm2
		punpcklwd mm2,mm0
		punpckhwd mm3,mm1
		paddd mm6,mm2
		paddd mm7,mm3
		add eax,16
		cmp eax,ecx
		jl xloop
 		add edi,edx
 		add esi,ebx
 		dec height
 		jnz yloop
		paddd mm6,mm7
		movq mm7,mm6
		psrlq mm6,32
		paddd mm6,mm7
		mov eax,diff
		movd [eax],mm6
 		emms
	}
}